import unittest
import os
import random
import string
from base64 import b64encode, b64decode
from komlogd.api import crypto, exceptions
from cryptography.hazmat.primitives.asymmetric import rsa

class ApiCryptoTest(unittest.TestCase):

    def test_load_private_key_failure_non_existent_file(self):
        ''' load_private_key should fail if filename does not exists '''
        filename='/a/path/to/nonexistentfile'
        with self.assertRaises(OSError) as cm:
            crypto.load_private_key(filename)
        self.assertEqual(cm.exception.errno, 2)

    def test_load_private_key_failure_filename_passed_is_not_a_file(self):
        ''' load_private_key should fail if filename passed is not a file '''
        filename='/tmp'
        self.assertTrue(os.path.isdir(filename))
        with self.assertRaises(OSError) as cm:
            crypto.load_private_key(filename)
        self.assertEqual(cm.exception.errno, 21)

    def test_load_private_key_failure_no_read_permission(self):
        ''' load_private_key should fail if we have no read permission over filename '''
        filename='/tmp/random_file_'+''.join(random.SystemRandom().choice(string.ascii_uppercase
            + string.digits) for _ in range(10))
        with os.fdopen(os.open(filename, os.O_WRONLY | os.O_CREAT, 0o200), 'w') as handle:
          handle.write('')
        with self.assertRaises(OSError) as cm:
            crypto.load_private_key(filename)
        os.remove(filename)
        self.assertEqual(cm.exception.errno, 13)

    def test_load_private_key_failure_invalid_key_file(self):
        ''' load_private_key should fail if we pass an invalid key file '''
        filename='/tmp/random_file_'+''.join(random.SystemRandom().choice(string.ascii_uppercase
            + string.digits) for _ in range(10))
        with os.fdopen(os.open(filename, os.O_WRONLY | os.O_CREAT, 0o600), 'w') as handle:
          handle.write('invalid_key')
        with self.assertRaises(exceptions.CryptoException) as cm:
            crypto.load_private_key(filename)
        os.remove(filename)

    def test_load_private_key_failure_public_key_filename_passed(self):
        ''' load_private_key should fail if we pass the public key file instead of the private one '''
        random_string=''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for _ in range(10))
        privkey_file=os.path.join('/tmp/',random_string+'.priv')
        pubkey_file=os.path.join('/tmp/',random_string+'.pub')
        privkey=crypto.generate_rsa_key()
        pubkey_generated=privkey.public_key()
        self.assertTrue(crypto.store_keys(privkey=privkey,privkey_file=privkey_file,pubkey_file=pubkey_file))
        with self.assertRaises(exceptions.CryptoException) as cm:
            loadedkey=crypto.load_private_key(pubkey_file)
        os.remove(privkey_file)
        os.remove(pubkey_file)

    def test_load_private_key_success(self):
        ''' load_private_key should succeed and return the private key '''
        random_string=''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for _ in range(10))
        privkey_file=os.path.join('/tmp/',random_string+'.priv')
        pubkey_file=os.path.join('/tmp/',random_string+'.pub')
        privkey=crypto.generate_rsa_key()
        pubkey_generated=privkey.public_key()
        self.assertTrue(crypto.store_keys(privkey=privkey,privkey_file=privkey_file,pubkey_file=pubkey_file))
        loadedkey=crypto.load_private_key(privkey_file)
        pubkey_loaded=privkey.public_key()
        os.remove(privkey_file)
        os.remove(pubkey_file)
        self.assertIsNotNone(loadedkey)
        self.assertEqual(crypto.serialize_public_key(pubkey_generated),crypto.serialize_public_key(pubkey_loaded))

    def test_load_public_key_failure_non_existent_file(self):
        ''' load_public_key should fail if filename does not exists '''
        filename='/a/path/to/nonexistentfile'
        with self.assertRaises(OSError) as cm:
            crypto.load_public_key(filename)
        self.assertEqual(cm.exception.errno, 2)

    def test_load_public_key_failure_filename_passed_is_not_a_file(self):
        ''' load_public_key should fail if filename passed is not a file '''
        filename='/tmp'
        self.assertTrue(os.path.isdir(filename))
        with self.assertRaises(OSError) as cm:
            crypto.load_public_key(filename)
        self.assertEqual(cm.exception.errno, 21)

    def test_load_public_key_failure_no_read_permission(self):
        ''' load_public_key should fail if we have no read permission over filename '''
        filename='/tmp/random_file_'+''.join(random.SystemRandom().choice(string.ascii_uppercase
            + string.digits) for _ in range(10))
        with os.fdopen(os.open(filename, os.O_WRONLY | os.O_CREAT, 0o200), 'w') as handle:
          handle.write('')
        with self.assertRaises(OSError) as cm:
            crypto.load_public_key(filename)
        os.remove(filename)
        self.assertEqual(cm.exception.errno, 13)

    def test_load_public_key_failure_invalid_key_file(self):
        ''' load_public_key should fail if we pass an invalid key file '''
        filename='/tmp/random_file_'+''.join(random.SystemRandom().choice(string.ascii_uppercase
            + string.digits) for _ in range(10))
        with os.fdopen(os.open(filename, os.O_WRONLY | os.O_CREAT, 0o600), 'w') as handle:
          handle.write('invalid_key')
        with self.assertRaises(exceptions.CryptoException) as cm:
            crypto.load_public_key(filename)
        os.remove(filename)

    def test_load_public_key_failure_private_key_filename_passed(self):
        ''' load_public_key should fail if we pass the private key file instead of the private one '''
        random_string=''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for _ in range(10))
        privkey_file=os.path.join('/tmp/',random_string+'.priv')
        pubkey_file=os.path.join('/tmp/',random_string+'.pub')
        privkey=crypto.generate_rsa_key()
        pubkey_generated=privkey.public_key()
        self.assertTrue(crypto.store_keys(privkey=privkey,privkey_file=privkey_file,pubkey_file=pubkey_file))
        with self.assertRaises(exceptions.CryptoException) as cm:
            loadedkey=crypto.load_public_key(privkey_file)
        os.remove(privkey_file)
        os.remove(pubkey_file)

    def test_load_public_key_success(self):
        ''' load_public_key should succeed and return the public key '''
        random_string=''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for _ in range(10))
        privkey_file=os.path.join('/tmp/',random_string+'.priv')
        pubkey_file=os.path.join('/tmp/',random_string+'.pub')
        privkey=crypto.generate_rsa_key()
        pubkey_generated=privkey.public_key()
        self.assertTrue(crypto.store_keys(privkey=privkey,privkey_file=privkey_file,pubkey_file=pubkey_file))
        pubkey_loaded=crypto.load_public_key(pubkey_file)
        os.remove(privkey_file)
        os.remove(pubkey_file)
        self.assertIsNotNone(pubkey_loaded)
        self.assertEqual(crypto.serialize_public_key(pubkey_generated),crypto.serialize_public_key(pubkey_loaded))

    def test_serialize_public_key_failure_invalid_key(self):
        ''' serialize_public_key should fail if key parameter is not a valid public key '''
        key='invalid_key'
        self.assertIsNone(crypto.serialize_public_key(key))

    def test_serialize_public_key_failure_private_key_passed(self):
        ''' serialize_public_key should fail if key parameter is not the public key but the private one '''
        privkey=crypto.generate_rsa_key()
        self.assertIsNone(crypto.serialize_public_key(privkey))

    def test_serialize_public_key_success(self):
        ''' serialize_public_key should succeed if we pass a valid public key '''
        privkey=crypto.generate_rsa_key()
        pubkey=privkey.public_key()
        self.assertIsNotNone(crypto.serialize_public_key(pubkey))

    def test_serialize_private_key_failure_invalid_key(self):
        ''' serialize_private_key should fail if key parameter is not a valid public key '''
        key='invalid_key'
        self.assertIsNone(crypto.serialize_private_key(key))

    def test_serialize_private_key_failure_public_key_passed(self):
        ''' serialize_private_key should fail if key parameter is not the private key but the public one '''
        privkey=crypto.generate_rsa_key()
        pubkey=privkey.public_key()
        self.assertIsNone(crypto.serialize_private_key(pubkey))

    def test_serialize_private_key_success(self):
        ''' serialize_private_key should succeed if we pass a valid private key '''
        privkey=crypto.generate_rsa_key()
        self.assertIsNotNone(crypto.serialize_private_key(privkey))

    def test_decrypt_failure_invalid_private_key(self):
        ''' decrypt should fail if private key is invalid '''
        privkey='privkey'
        ciphertext='ciphertext'.encode()
        self.assertIsNone(crypto.decrypt(privkey=privkey, ciphertext=ciphertext))

    def test_decrypt_failure_invalid_ciphertext(self):
        ''' decrypt should fail if ciphertext is not of type bytes '''
        privkey=crypto.generate_rsa_key()
        ciphertext='string type'
        self.assertIsNone(crypto.decrypt(privkey=privkey, ciphertext=ciphertext))

    def test_decrypt_success(self):
        ''' decrypt should succeed '''
        privkey=crypto.generate_rsa_key()
        pubkey=privkey.public_key()
        text='text to encrypt'.encode()
        ciphertext=crypto.encrypt(pubkey,text)
        self.assertEqual(crypto.decrypt(privkey=privkey, ciphertext=ciphertext), text)

    def test_encrypt_failure_invalid_pubkey(self):
        ''' encrypt should fail if pubkey is not valid '''
        pubkey='pubkey'
        plaintext='text to encrypt'.encode()
        self.assertIsNone(crypto.encrypt(pubkey,plaintext))

    def test_encrypt_failure_invalid_plaintext(self):
        ''' encrypt should fail if pubkey is not valid bytes encoded text '''
        privkey=crypto.generate_rsa_key()
        pubkey=privkey.public_key()
        plaintext='text to encrypt is a string'
        self.assertIsNone(crypto.encrypt(pubkey,plaintext))

    def test_encrypt_success(self):
        ''' encrypt should succeed '''
        privkey=crypto.generate_rsa_key()
        pubkey=privkey.public_key()
        plaintext='text to encrypt is bytes encoded'.encode()
        self.assertIsNotNone(crypto.encrypt(pubkey,plaintext))

    def test_sign_message_failure_invalid_privkey(self):
        ''' sign_message should return None if privkey is invalid '''
        privkey='privkey'
        message=b64encode('message'.encode()).decode()
        self.assertIsNone(crypto.sign_message(privkey,message))

    def test_sign_message_failure_invalid_message(self):
        ''' sign_message should return None if message is not a string '''
        privkey=crypto.generate_rsa_key()
        message=234234234234
        self.assertIsNone(crypto.sign_message(privkey,message))

    def test_sign_message_success(self):
        ''' sign_message should return the message signed '''
        privkey=crypto.generate_rsa_key()
        message=b64encode('message'.encode()).decode()
        signed=crypto.sign_message(privkey, message)
        self.assertIsNotNone(signed)
        self.assertTrue(isinstance(signed, str))

    def test_get_hash_failure_invalid_message(self):
        ''' get_hash should return None if message is not a bytes variable '''
        message='message'
        self.assertIsNone(crypto.get_hash(message))

    def test_get_hash_success(self):
        ''' get_hash should return the hash of the message passed '''
        message='message'.encode()
        self.assertIsNotNone(crypto.get_hash(message))

    def test_process_challenge_failure_invalid_challenge(self):
        ''' process_challenge should return None if challenge received cannot be decrypted '''
        privkey=crypto.generate_rsa_key()
        challenge=b64encode('challenge'.encode()).decode()
        self.assertIsNone(crypto.process_challenge(privkey, challenge))

    def test_generate_rsa_key_success(self):
        ''' generate_rsa_key should return a RSAPrivateKey object with size 4096 '''
        privkey=crypto.generate_rsa_key()
        self.assertTrue(isinstance(privkey, rsa.RSAPrivateKey))
        self.assertEqual(privkey.key_size, 4096)

    def test_store_keys_failure_non_private_key_passed(self):
        ''' store_keys should fail if privkey is not a valid key '''
        privkey='privkey'
        privkey_file='/tmp/test_store_keys_failure_non_private_key_passed.priv'
        pubkey_file='/tmp/test_store_keys_failure_non_private_key_passed.pub'
        self.assertFalse(crypto.store_keys(privkey, privkey_file, pubkey_file))

    def test_store_keys_failure_no_permission_to_create_private_file(self):
        ''' store_keys should fail if process has no permission to create private file '''
        privkey=crypto.generate_rsa_key()
        privkey_file='/root/test_store_keys_failure_no_permission_to_create_private_file.priv'
        pubkey_file='/tmp/test_store_keys_failure_permission_to_create_private_file.pub'
        self.assertFalse(crypto.store_keys(privkey, privkey_file, pubkey_file))
        self.assertFalse(os.path.isfile(privkey_file))
        self.assertFalse(os.path.isfile(pubkey_file))

    def test_store_keys_failure_no_permission_to_create_public_file(self):
        ''' store_keys should fail if process has no permission to create public file '''
        privkey=crypto.generate_rsa_key()
        privkey_file='/tmp/test_store_keys_failure_permission_to_create_public_file.priv'
        pubkey_file='/root/test_store_keys_failure_permission_to_create_public_file.pub'
        self.assertFalse(crypto.store_keys(privkey, privkey_file, pubkey_file))
        self.assertFalse(os.path.isfile(privkey_file))
        self.assertFalse(os.path.isfile(pubkey_file))

    def test_get_printable_pubkey_failure_invalid_pubkey(self):
        ''' get_printable_pubkey should return None if pubkey is invalid '''
        pubkey='pubkey'
        self.assertIsNone(crypto.get_printable_pubkey(pubkey))

    def test_get_printable_pubkey_success(self):
        ''' get_printable_pubkey should return the string serialization of the public key '''
        privkey=crypto.generate_rsa_key()
        pubkey=privkey.public_key()
        self.assertIsNotNone(crypto.get_printable_pubkey(pubkey))




import unittest
import uuid
import asyncio
import pandas as pd
from komlogd.api import transfer_methods, exceptions
from komlogd.api.model.store import MetricsStore
from komlogd.api.protocol.model.types import Metric, Datasource, Datapoint, Sample
from komlogd.api.protocol.model.schedules import OnUpdateSchedule
from komlogd.api.protocol.model.transfer_methods import DataRequirements

class ApiTransferMethodsTest(unittest.TestCase):

    def test_transfermethod_failure_invalid_p_in(self):
        ''' creation of a transfermethod object should fail if p_in is invalid '''
        p_ins = [1,'str',{'set'},['list'],('tupl','e'),Metric(uri='uri'),pd.Timestamp('now')]
        for p_in in p_ins:
            with self.assertRaises(exceptions.BadParametersException) as cm:
                tm=transfer_methods.transfermethod(p_in=p_in)
            self.assertEqual(cm.exception.msg, '"p_in" attribute must be a dict')

    def test_transfermethod_failure_reserved_p_in_key(self):
        ''' creation of a transfermethod object should fail if p_in has a reserved key '''
        p_ins=[{'ts':1},{'updated':1},{'others':1}]
        for p_in in p_ins:
            key = list(p_in.keys())[0]
            with self.assertRaises(exceptions.BadParametersException) as cm:
                tm=transfer_methods.transfermethod(p_in=p_in)
            self.assertEqual(cm.exception.msg, 'Invalid input parameter. "{}" is a reserved parameter'.format(key))

    def test_transfermethod_failure_invalid_p_out(self):
        ''' creation of a transfermethod object should fail if p_out is invalid '''
        p_outs = [1,'str',{'set'},['list'],('tupl','e'),Metric(uri='uri'),pd.Timestamp('now')]
        for p_out in p_outs:
            with self.assertRaises(exceptions.BadParametersException) as cm:
                tm=transfer_methods.transfermethod(p_out=p_out)
            self.assertEqual(cm.exception.msg, '"p_out" attribute must be a dict')

    def test_transfermethod_failure_reserved_p_out_key(self):
        ''' creation of a transfermethod object should fail if p_out has a reserved key '''
        p_outs=[{'ts':1},{'updated':1},{'others':1}]
        for p_out in p_outs:
            key = list(p_out.keys())[0]
            with self.assertRaises(exceptions.BadParametersException) as cm:
                tm=transfer_methods.transfermethod(p_out=p_out)
            self.assertEqual(cm.exception.msg, 'Invalid output parameter. "{}" is a reserved parameter'.format(key))

    def test_transfermethod_failure_invalid_schedule(self):
        ''' creation of a transfermethod object should fail if shedule parameter is invalid '''
        schedule='once upon a time'
        with self.assertRaises(exceptions.BadParametersException) as cm:
            tm=transfer_methods.transfermethod(schedule=schedule)
        self.assertEqual(cm.exception.msg, 'Invalid "schedule" attribute')

    def test_transfermethod_failure_invalid_data_reqs(self):
        ''' creation of a transfermethod object should fail if data_reqs parameter is invalid '''
        data_reqs='give me all'
        with self.assertRaises(exceptions.BadParametersException) as cm:
            tm=transfer_methods.transfermethod(data_reqs=data_reqs)
        self.assertEqual(cm.exception.msg, 'Invalid data_reqs parameter')

    def test_transfermethod_success_no_metrics_in_params(self):
        ''' creation of a transfermethod object should succeed if uri is valid '''
        p_in={'arg1':'valid.uri','arg2':'another.uri'}
        p_out={'arg3':'valid.uri','arg4':'another.uri'}
        data_reqs = DataRequirements(past_delta=pd.Timedelta('6h'))
        tm=transfer_methods.transfermethod(p_in=p_in, p_out=p_out, data_reqs=data_reqs, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.schedule, None)
        self.assertEqual(tm.data_reqs, data_reqs)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.allow_loops, True)

    def test_transfermethod_success_registering_transfermethod_no_metrics_in_params(self):
        '''transfermethod object should be able to register the associated method successfully '''
        p_in={'arg1':'valid.uri','arg2':'another.uri'}
        p_out={'arg3':'valid.uri','arg4':'another.uri'}
        data_reqs = DataRequirements(past_delta=pd.Timedelta('6h'))
        tm=transfer_methods.transfermethod(p_in=p_in, p_out=p_out, data_reqs=data_reqs, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.schedule, None)
        self.assertEqual(tm.data_reqs, data_reqs)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.allow_loops, True)
        def func():
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(tm._func_params,{})
        self.assertNotEqual(tm.schedule, None)
        self.assertTrue(isinstance(tm.schedule, OnUpdateSchedule))
        self.assertEqual(tm.schedule.metrics, [])
        self.assertEqual(tm.schedule.exec_on_load, False)
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_transfermethod_success_registering_transfermethod_with_input_params(self):
        '''transfermethod object should be able to register the associated method and find input params '''
        p_in={'arg1':Datapoint('valid.datapoint'),'arg2':Datasource('valid.datasource')}
        data_reqs = DataRequirements(past_delta=pd.Timedelta('6h'))
        tm=transfer_methods.transfermethod(p_in=p_in, data_reqs=data_reqs, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, data_reqs)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, {})
        self.assertEqual(tm.allow_loops, True)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_in_routes,{'arg1':[[('s',0)]],'arg2':[[('s',0)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datapoint('valid.datapoint'),Datasource('valid.datasource')})
        self.assertNotEqual(tm.schedule, None)
        self.assertTrue(isinstance(tm.schedule, OnUpdateSchedule))
        self.assertEqual(sorted(tm.schedule.metrics, key=lambda x:x.uri), sorted(list(p_in.values()),key=lambda x:x.uri))
        self.assertEqual(tm.schedule.exec_on_load, False)
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_transfermethod_success_registering_transfermethod_onupdateschedule_different_input_metric(self):
        '''transfermethod object should be able to register the associated method and find input params '''
        p_in={'arg1':Datapoint('valid.datapoint'),'arg2':Datasource('valid.datasource')}
        data_reqs = DataRequirements(past_delta=pd.Timedelta('6h'))
        schedule=OnUpdateSchedule(metrics=[Datapoint('valid.datapoint2')], exec_on_load=True)
        tm=transfer_methods.transfermethod(p_in=p_in, data_reqs=data_reqs, schedule=schedule, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, data_reqs)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, {})
        self.assertEqual(tm.allow_loops, True)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_in_routes,{'arg1':[[('s',0)]],'arg2':[[('s',0)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datapoint('valid.datapoint'),Datasource('valid.datasource')})
        self.assertNotEqual(tm.schedule, None)
        self.assertTrue(isinstance(tm.schedule, OnUpdateSchedule))
        self.assertEqual(tm.schedule.metrics, schedule.metrics)
        self.assertEqual(tm.schedule.exec_on_load, True)
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_transfermethod_success_registering_transfermethod_with_output_params(self):
        '''transfermethod object should be able to register the associated method and find output params '''
        p_out={'arg1':Datapoint('valid.datapoint'),'arg2':Datasource('valid.datasource')}
        data_reqs = DataRequirements(past_delta=pd.Timedelta('6h'))
        tm=transfer_methods.transfermethod(p_out=p_out, data_reqs=data_reqs, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, data_reqs)
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.p_in, {})
        self.assertEqual(tm.allow_loops, True)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_out_routes,{'arg1':[[('s',0)]],'arg2':[[('s',0)]]})
        self.assertEqual(tm._p_in_routes,{})
        self.assertEqual(tm._m_in, set())
        self.assertNotEqual(tm.schedule, None)
        self.assertTrue(isinstance(tm.schedule, OnUpdateSchedule))
        self.assertEqual(tm.schedule.metrics, [])
        self.assertEqual(tm.schedule.exec_on_load, False)
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_inspect_input_params_success_list(self):
        '''transfermethod object should be able to identify params in a list '''
        p_in={'arg1':[Datapoint('valid.datapoint'), Datasource('valid.datasource')]}
        tm=transfer_methods.transfermethod(p_in=p_in)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_out, {})
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.allow_loops, False)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_in_routes,{'arg1':[[('i',0)],[('i',1)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datapoint('valid.datapoint'),Datasource('valid.datasource')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_inspect_input_params_success_dict(self):
        '''transfermethod object should be able to identify params in a dict '''
        p_in={'arg1':{'dp':Datapoint('valid.datapoint'), 'ds':Datasource('valid.datasource')}}
        tm=transfer_methods.transfermethod(p_in=p_in)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_out, {})
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.allow_loops, False)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_in_routes,{'arg1':[[('k','dp')],[('k','ds')]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datapoint('valid.datapoint'),Datasource('valid.datasource')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_inspect_input_params_success_class(self):
        '''transfermethod object should be able to identify params in a user defined class instance '''
        class MyObj:
            def __init__(self, uri):
                self.uri = uri
                self.ds = Datasource('.'.join((uri,'ds')))
                self.dp = Datapoint('.'.join((uri,'dp')))
        p_in={'arg1':MyObj('valid')}
        tm=transfer_methods.transfermethod(p_in=p_in)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_out, {})
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.allow_loops, False)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_in_routes,{'arg1':[[('a','__dict__'),('k','ds')],[('a','__dict__'),('k','dp')]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datasource('valid.ds'),Datapoint('valid.dp')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_inspect_input_params_success_combination(self):
        '''transfermethod object should be able to identify params in a user defined class instance '''
        class MyObj:
            def __init__(self, uri):
                self.uri = uri
                self.elements={'metrics':[Datasource('.'.join((uri,'ds'))), Datapoint('.'.join((uri,'dp')))]}
        p_in={'arg1':MyObj('valid')}
        tm=transfer_methods.transfermethod(p_in=p_in)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_out, {})
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.allow_loops, False)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_in_routes,{'arg1':[[('a','__dict__'),('k','elements'),('k','metrics'),('i',0)],[('a','__dict__'),('k','elements'),('k','metrics'),('i',1)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datasource('valid.ds'),Datapoint('valid.dp')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_inspect_output_params_success_combination(self):
        '''transfermethod object should be able to identify params in a user defined class instance '''
        class MyObj:
            def __init__(self, uri):
                self.uri = uri
                self.elements={'metrics':[Datasource('.'.join((uri,'ds'))), Datapoint('.'.join((uri,'dp')))]}
        p_out={'arg1':MyObj('valid')}
        tm=transfer_methods.transfermethod(p_out=p_out)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_in, {})
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.allow_loops, False)
        def func(arg1, arg2):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['arg1','arg2'])
        self.assertEqual(tm._p_out_routes,{'arg1':[[('a','__dict__'),('k','elements'),('k','metrics'),('i',0)],[('a','__dict__'),('k','elements'),('k','metrics'),('i',1)]]})
        self.assertEqual(tm._p_in_routes,{})
        self.assertEqual(tm._m_in, set())
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))

    def test_process_exec_result_allow_loops(self):
        '''transfermethod _process_exec_result should allow modifying input parameters if loops are allowed '''
        class MySession:
            def __init__(self):
                self._metrics_store = MetricsStore(owner='user')
            async def send_samples(self, samples):
                self.samples = samples
        p_in = {'ds':Datasource('valid.ds')}
        p_out={'ds':None} #if arg is on p_in, no matter what we put here, only checks that the key is on p_in too
        tm=transfer_methods.transfermethod(p_in=p_in, p_out=p_out, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.allow_loops, True)
        def func(ds):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['ds'])
        self.assertEqual(tm._p_in_routes,{'ds':[[('s',0)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datasource('valid.ds')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))
        my_session = MySession()
        ds = Datasource('valid.ds')
        ds.data = pd.Series('ds content', index=[pd.Timestamp('now',tz='utc')])
        params = {'ds':ds}
        expected_sample = Sample(metric=ds, ts=ds.data.index[0], data=ds.data[0])
        loop=asyncio.get_event_loop()
        loop.run_until_complete(tm._process_exec_result(session=my_session, params=params))
        self.assertEqual(len(my_session.samples),1)
        self.assertEqual(my_session.samples[0].metric, ds)
        self.assertEqual(my_session.samples[0].ts, expected_sample.ts)
        self.assertEqual(my_session.samples[0].data, expected_sample.data)

    def test_process_exec_result_do_not_allow_loops(self):
        '''transfermethod _process_exec_result should ignore modifying input parameters if loops aren't allowed '''
        class MySession:
            def __init__(self):
                self._metrics_store = MetricsStore(owner='user')
            async def send_samples(self, samples):
                self.samples = samples
        p_in = {'ds':Datasource('valid.ds')}
        p_out={'ds':None} #if arg is on p_in, no matter what we put here, only checks that the key is on p_in too
        tm=transfer_methods.transfermethod(p_in=p_in, p_out=p_out)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.allow_loops, False)
        def func(ds):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['ds'])
        self.assertEqual(tm._p_in_routes,{'ds':[[('s',0)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datasource('valid.ds')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))
        my_session = MySession()
        ds = Datasource('valid.ds')
        ds.data = pd.Series('ds content', index=[pd.Timestamp('now',tz='utc')])
        params = {'ds':ds}
        loop=asyncio.get_event_loop()
        loop.run_until_complete(tm._process_exec_result(session=my_session, params=params))
        self.assertEqual(len(my_session.samples),0)


    def test_process_exec_result_allow_loops_but_existing_in_store(self):
        '''transfermethod _process_exec_result should not send already existing samples in store '''
        class MySession:
            def __init__(self):
                self._metrics_store = MetricsStore(owner='user')
            async def send_samples(self, samples):
                self.samples = samples
        p_in = {'ds':Datasource('valid.ds')}
        p_out={'ds':None} #if arg is on p_in, no matter what we put here, only checks that the key is on p_in too
        tm=transfer_methods.transfermethod(p_in=p_in, p_out=p_out, allow_loops=True)
        self.assertTrue(isinstance(tm.mid,uuid.UUID))
        self.assertEqual(tm.last_exec, None)
        self.assertEqual(tm.data_reqs, None)
        self.assertEqual(tm.p_in, p_in)
        self.assertEqual(tm.p_out, p_out)
        self.assertEqual(tm.allow_loops, True)
        def func(ds):
            pass
        f=tm(func)
        self.assertEqual(f,func)
        self.assertEqual(list(tm._func_params.keys()),['ds'])
        self.assertEqual(tm._p_in_routes,{'ds':[[('s',0)]]})
        self.assertEqual(tm._p_out_routes,{})
        self.assertEqual(tm._m_in, {Datasource('valid.ds')})
        self.assertIsNotNone(getattr(tm,'f',None))
        self.assertTrue(asyncio.iscoroutinefunction(tm.f))
        my_session = MySession()
        ds = Datasource('valid.ds')
        ds.data = pd.Series('ds content', index=[pd.Timestamp('now',tz='utc')])
        params = {'ds':ds}
        my_session._metrics_store.store(metric=ds, ts=ds.data.index[0], content=ds.data[0])
        loop=asyncio.get_event_loop()
        loop.run_until_complete(tm._process_exec_result(session=my_session, params=params))
        self.assertEqual(len(my_session.samples),0)




import unittest
import decimal
import pandas as pd
import datetime
import uuid
from komlogd.api.model import store
from komlogd.api.protocol.model.types import Datasource, Datapoint
from komlogd.api.protocol.model.transfer_methods import DataRequirements

class ApiModelStoreTest(unittest.TestCase):

    def test_store_creation_success(self):
        ''' store object creation without parameters should succeed '''
        ms=store.MetricsStore(owner='owner')
        self.assertTrue(isinstance(ms, store.MetricsStore))
        self.assertEqual(ms.owner, 'owner')
        self.assertEqual(ms.maintenance_exec_delta.seconds, 300)
        self.assertEqual(ms.default_reqs.past_delta.seconds, 600)
        self.assertEqual(ms.default_reqs.past_count, 2)
        self.assertEqual(ms._metric_reqs, {})
        self.assertEqual(ms._series, {})

    def test_store_creation_with_parameters_success(self):
        ''' store object creation with parameters should succeed '''
        purge_exec_delta=pd.Timedelta('1 d')
        reqs=DataRequirements(past_delta=pd.Timedelta('24 h'))
        ms=store.MetricsStore(owner='owner', maintenance_exec_delta=purge_exec_delta, data_delta=reqs)
        self.assertTrue(isinstance(ms, store.MetricsStore))
        self.assertTrue(isinstance(ms._series, dict))
        self.assertEqual(ms.maintenance_exec_delta.seconds, 0)
        self.assertEqual(ms.maintenance_exec_delta.days, 1)
        self.assertEqual(ms.default_reqs.past_delta.seconds, 0)
        self.assertEqual(ms.default_reqs.past_delta.days, 1)
        self.assertEqual(ms._metric_reqs, {})

    def test_store_new_datasource_metric_content_success(self):
        ''' if we store a new ds metric in the store, the series should be created with the new value and ts '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='ds.uri')
        content='datasource content with 1 value.'
        ts=pd.Timestamp('now',tz='utc')
        ms.store(metric=metric, ts=ts, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(serie.index[0],ts)
        self.assertEqual(serie[0],content)
        self.assertEqual(ms._metric_reqs, {})

    def test_store_new_datasource_metric_content_failure_invalid_ds_content(self):
        ''' storing a datasource metric with invalid content should fail '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='ds.uri')
        content=['a','list','4','example']
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            ms.store(metric=metric, ts=ts, content=content)
        self.assertEqual(str(cm.exception),'content is not a string')

    def test_store_new_datapoint_metric_content_success(self):
        ''' if we store a new dp metric in the store, the series should be created with the new value and ts '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts=pd.Timestamp('now',tz='utc')
        ms.store(metric=metric, ts=ts, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(serie.index[0],ts)
        self.assertEqual(serie[0],float(content))
        self.assertEqual(ms._metric_reqs, {})

    def test_store_new_datapoint_metric_content_failure_invalid_dp_content(self):
        ''' storing a datapoint metric with invalid content should fail '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content='content'
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            ms.store(metric=metric, ts=ts, content=content)
        self.assertEqual(str(cm.exception),'datapoint value not valid')

    def test_store_new_metric_content_failure_ts_has_no_timezone(self):
        ''' Timestamps without timezone information are not valid '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts=pd.Timestamp('now')
        with self.assertRaises(TypeError) as cm:
            ms.store(metric=metric, ts=ts, content=content)
        self.assertEqual(str(cm.exception),'timezone is required')

    def test_store_already_existing_datapoint_metric_content_success(self):
        ''' if we store an already existing dp metric in the store, contents should be appended to the existing metric '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35')
        ts=pd.Timestamp('now',tz='utc')
        ms.store(metric=metric, ts=ts, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(serie.index[0],ts)
        self.assertEqual(serie[0],int(content))
        content2=decimal.Decimal('36.221')
        ts2=pd.Timestamp('now',tz='utc')
        ms.store(metric=metric, ts=ts2, content=content2)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),2)
        self.assertEqual(serie.index[1],ts2)
        self.assertEqual(serie[1],float(content2))

    def test_overwrite_existing_datapoint_content(self):
        ''' if we store an already existing ts, the content will be replaced '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts=pd.Timestamp('now',tz='utc')
        ms.store(metric=metric, ts=ts, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),1)
        self.assertEqual(serie.index[0],ts)
        self.assertEqual(serie[0],float(content))
        content2=decimal.Decimal('36.221')
        ts2=ts
        ms.store(metric=metric, ts=ts2, content=content2)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),1)
        self.assertEqual(serie.index[0],ts)
        self.assertEqual(serie[0],float(content2))

    def test_sort_index_if_ts_less_than_last_in_index(self):
        ''' if we insert a row which a ts older than the last in the series, the index should be sorted '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts1=pd.Timestamp('2016-07-30',tz='utc')
        ts2=pd.Timestamp('2016-07-31',tz='utc')
        ts3=pd.Timestamp('2016-07-29',tz='utc')
        ms.store(metric=metric, ts=ts1, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),1)
        self.assertEqual(serie.index[0],ts1)
        ms.store(metric=metric, ts=ts2, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),2)
        self.assertEqual(serie.index[0],ts1)
        self.assertEqual(serie.index[1],ts2)
        ms.store(metric=metric, ts=ts3, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),3)
        self.assertEqual(serie.index[0],ts3)
        self.assertEqual(serie.index[1],ts1)
        self.assertEqual(serie.index[2],ts2)

    def test_purge_should_delete_contents_older_than_set_limit(self):
        ''' purge should delete contents older than limit set by param data_delta. '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts3=pd.Timestamp('now',tz='utc')
        ts2=ts3 - pd.Timedelta('1s')
        ts1=ts2 - pd.Timedelta('1s')
        ms.store(metric=metric, ts=ts1, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),1)
        self.assertEqual(serie.index[0],ts1)
        ms.store(metric=metric, ts=ts2, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),2)
        self.assertEqual(serie.index[0],ts1)
        self.assertEqual(serie.index[1],ts2)
        ms.store(metric=metric, ts=ts3, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),3)
        self.assertEqual(serie.index[0],ts1)
        self.assertEqual(serie.index[1],ts2)
        self.assertEqual(serie.index[2],ts3)
        ms.purge()
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),3) #DataRequirements by default are 2 samples or 10 min of data
        ms.default_reqs=DataRequirements(past_delta=pd.Timedelta('1 us'))
        ms.purge()
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),0)

    def test_purge_should_not_execute_because_last_purge_exec(self):
        ''' purge should not execute if last_purge_exec + purge_exec_delta is a time in the future '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts1=pd.Timestamp('1990-03-01',tz='utc')
        ts2=pd.Timestamp('1991-03-01',tz='utc')
        ts3=pd.Timestamp('1992-03-01',tz='utc')
        ms.store(metric=metric, ts=ts1, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),1)
        self.assertEqual(serie.index[0],ts1)
        ms.store(metric=metric, ts=ts2, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),2)
        self.assertEqual(serie.index[0],ts1)
        self.assertEqual(serie.index[1],ts2)
        ms.store(metric=metric, ts=ts3, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),3)
        self.assertEqual(serie.index[0],ts1)
        self.assertEqual(serie.index[1],ts2)
        self.assertEqual(serie.index[2],ts3)
        ms.purge()
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),2) #DataRequirements by default are 2 samples or 10 min of data
        ms.default_reqs=DataRequirements(past_delta=pd.Timedelta('1 us'))
        ms.purge()
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),0)

    def test_purge_should_take_specific_metric_delta_if_it_exists(self):
        ''' purge should take specific metric delta if it exists '''
        reqs=DataRequirements(past_delta=pd.Timedelta('24h'))
        ms=store.MetricsStore(owner='owner',maintenance_exec_delta=pd.Timedelta('1 us'),data_delta=reqs)
        metric=Datapoint(uri='dp.uri')
        content=decimal.Decimal('35.221')
        ts1=pd.Timestamp('1990-03-01',tz='utc')
        ts2=pd.Timestamp('1991-03-01',tz='utc')
        ts3=pd.Timestamp('1992-03-01',tz='utc')
        ms.store(metric=metric, ts=ts1, content=content)
        ms.store(metric=metric, ts=ts2, content=content)
        ms.store(metric=metric, ts=ts3, content=content)
        global_metric=Datapoint(uri='dp.uri2')
        ms.store(metric=global_metric, ts=ts1, content=content)
        ms.store(metric=global_metric, ts=ts2, content=content)
        ms.store(metric=global_metric, ts=ts3, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),3)
        serie = ms.get_serie(global_metric)
        self.assertEqual(len(serie),3)
        specific_delta=DataRequirements(past_delta=pd.Timestamp('now',tz='utc')-pd.Timestamp('1991-02-28',tz='utc'))
        self.assertTrue(ms.add_metric_data_reqs(metric, specific_delta))
        ms.purge()
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie),2)
        serie = ms.get_serie(global_metric)
        self.assertEqual(len(serie),0)

    def test_add_metric_data_reqs_success_non_existent_metric_previously(self):
        ''' add_metric_data_reqs should succeed seting the reqs to a non previously existent metric '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        delta=pd.Timedelta('48 h')
        reqs = DataRequirements(past_delta=delta)
        self.assertTrue(ms.add_metric_data_reqs(metric=metric, reqs=reqs))
        d_stablished=ms.get_metric_data_reqs(metric)
        self.assertEqual(d_stablished.past_delta, delta)

    def test_add_metric_data_reqs_success_metric_existed_previously_past_delta_set(self):
        ''' add_metric_data_reqs should succeed seting the data delta to a previously existent metric if delta is greater than the existent one '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        delta=pd.Timedelta('48 h')
        reqs = DataRequirements(past_delta=delta)
        self.assertTrue(ms.add_metric_data_reqs(metric=metric, reqs=reqs))
        d_stablished=ms.get_metric_data_reqs(metric)
        self.assertEqual(d_stablished.past_delta, delta)
        new_delta=pd.Timedelta('96 h')
        reqs = DataRequirements(past_delta=new_delta)
        self.assertTrue(ms.add_metric_data_reqs(metric=metric, reqs=reqs))
        d_stablished=ms.get_metric_data_reqs(metric)
        self.assertEqual(d_stablished.past_delta, new_delta)

    def test_add_metric_data_delta_success_metric_existed_previously_past_delta_smaller(self):
        ''' add_metric_data_reqs should succeed seting the data delta to a previously existent metric if delta is smaller than the existent one '''
        ms=store.MetricsStore(owner='owner')
        metric=Datapoint(uri='dp.uri')
        delta=pd.Timedelta('48 h')
        reqs = DataRequirements(past_delta=delta)
        self.assertTrue(ms.add_metric_data_reqs(metric=metric, reqs=reqs))
        d_stablished=ms.get_metric_data_reqs(metric)
        self.assertEqual(d_stablished.past_delta, delta)
        new_delta=pd.Timedelta('24 h')
        reqs = DataRequirements(past_delta=new_delta)
        self.assertTrue(ms.add_metric_data_reqs(metric=metric, reqs=reqs))
        d_stablished=ms.get_metric_data_reqs(metric)
        self.assertEqual(d_stablished.past_delta, delta)

    def test_get_metric_data_reqs_success_no_delta_found(self):
        ''' get_metric_data_reqs should return None if metric has no delta set '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='ds.uri')
        self.assertIsNone(ms.get_metric_data_reqs(metric))

    def test_get_metric_data_reqs_success_delta_found(self):
        ''' get_metric_data_reqs should return the delta set if metric has it '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='ds.uri')
        delta=pd.Timedelta('48 h')
        reqs = DataRequirements(past_delta=delta)
        self.assertTrue(ms.add_metric_data_reqs(metric=metric, reqs=reqs))
        d_stablished=ms.get_metric_data_reqs(metric)
        self.assertEqual(d_stablished.past_delta, delta)

    def test_get_serie_metric_not_found(self):
        ''' get_serie should return an empty serie if metric is not found '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='ds.uri')
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie), 0)

    def test_get_serie_full_data(self):
        ''' get_serie should return a copy of the full serie if no interval or count parameter is passed'''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='owner:ds.uri')
        tss = [pd.Timestamp(value,tz='utc') for value in [ns*1000 for ns in range(1,1000)]]
        for ts in tss:
            content = str(ts)
            ms.store(metric=metric, ts=ts, content=content)
        serie = ms.get_serie(metric)
        self.assertEqual(len(serie), len(ms._series[metric.uri]))

    def test_get_serie_data_interval(self):
        ''' get_serie should return a copy of the interval passed '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='owner:ds.uri')
        tss = [pd.Timestamp(value,tz='utc') for value in [ns*1000 for ns in range(1,1000)]]
        for ts in tss:
            content = str(ts)
            ms.store(metric=metric, ts=ts, content=content)
        ets=pd.Timestamp(500000, tz='utc')
        its=pd.Timestamp(1000, tz='utc')
        serie = ms.get_serie(metric, ets=ets, its=its)
        self.assertEqual(len(serie), 500)
        self.assertEqual(serie.index[0],its)
        self.assertEqual(serie.ix[0],str(its))
        self.assertEqual(serie.index[-1],ets)
        self.assertEqual(serie.ix[-1],str(ets))

    def test_get_serie_data_count(self):
        ''' get_serie should return a copy of the serie with as much rows as requested '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='owner:ds.uri')
        tss = [pd.Timestamp(value,tz='utc') for value in [ns*1000 for ns in range(1,1000)]]
        for ts in tss:
            content = str(ts)
            ms.store(metric=metric, ts=ts, content=content)
        ets=pd.Timestamp(500000, tz='utc')
        count=10
        serie = ms.get_serie(metric, ets=ets, count=count)
        self.assertEqual(len(serie), 10)
        self.assertEqual(serie.index[-1],ets)
        self.assertEqual(serie.ix[-1],str(ets))

    def test_get_serie_data_last(self):
        ''' get_serie should return a copy of the serie with the last row with ts older than the requested one '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='owner:ds.uri')
        tss = [pd.Timestamp(value,tz='utc') for value in [ns*1000 for ns in range(1,1000)]]
        for ts in tss:
            content = str(ts)
            ms.store(metric=metric, ts=ts, content=content)
        ets_req=pd.Timestamp(500500, tz='utc')
        ets_data=pd.Timestamp(500000, tz='utc')
        serie = ms.get_serie(metric, ets=ets_req)
        self.assertEqual(len(serie), 1)
        self.assertEqual(serie.index[-1],ets_data)
        self.assertEqual(serie.ix[-1],str(ets_data))

    def test_get_serie_modify_data_does_not_modify_store(self):
        ''' get_serie should return a copy of the serie '''
        ms=store.MetricsStore(owner='owner')
        metric=Datasource(uri='owner:ds.uri')
        tss = [pd.Timestamp(value,tz='utc') for value in [ns*1000 for ns in range(1,1000)]]
        for ts in tss:
            content = str(ts)
            ms.store(metric=metric, ts=ts, content=content)
        ets=pd.Timestamp(500000, tz='utc')
        its=pd.Timestamp(1000, tz='utc')
        serie = ms.get_serie(metric, ets=ets, its=its)
        self.assertEqual(len(serie), 500)
        self.assertEqual(serie.index[0],its)
        self.assertEqual(serie.ix[0],str(its))
        self.assertEqual(serie.index[-1],ets)
        self.assertEqual(serie.ix[-1],str(ets))
        for i in range(0,500):
            serie[i] = str(serie[i])+'modified'
        serie2 = ms.get_serie(metric, ets=ets, its=its)
        for i in range(0,500):
            self.assertEqual(serie[i],str(pd.Timestamp((i+1)*1000, tz='utc'))+'modified')
            self.assertEqual(serie2[i],str(pd.Timestamp((i+1)*1000, tz='utc')))




import pandas as pd
from komlogd.api.protocol.model import validation
from komlogd.api.protocol.model.types import Metric

def get_global_uri(metric, owner):
    if not isinstance(metric, Metric):
        raise TypeError('Invalid metric')
    if validation.is_local_uri(metric.uri):
        return ':'.join((owner.lower(),metric.uri))
    elif validation.is_global_uri(metric.uri):
        owner,local_uri=metric.uri.split(':')
        uri = ':'.join((owner.lower(),local_uri))
        return uri
    else:
        raise TypeError('Invalid metric uri')




KOMLOG_USERNAME='username'
KOMLOG_KEYFILE='keyfile'
ENTRY_LOG='logging'
ENTRY_TRANSFERS='transfers'
LOG_FILE='filename'
LOG_DIR='dirname'
LOG_LEVEL='level'
LOG_ROTATION='rotation'
LOG_MAX_BYTES='max_bytes'
LOG_BACKUP_COUNT='backup_count'
LOG_FORMAT='format'
TRANSFERS_ENABLED='enabled'
TRANSFERS_FILE='filename'




import asyncio
import traceback
from komlogd.api import logging

class ExitMessage:
    pass

class AsyncQueue:
    def __init__(self, num_workers, on_msg, name, loop=None):
        loop = loop or asyncio.get_event_loop()
        self._loop = loop
        self._num_workers = num_workers
        self._on_msg = on_msg
        self._name = name
        self._queue = asyncio.Queue()
        self._workers = None
        self._timeout = 0

    def start(self):
        logging.logger.debug('Starting AsyncQueue {}'.format(self._name))
        assert self._workers is None
        self._workers = [asyncio.ensure_future(self._worker_loop(i+1), loop=self._loop) for i in range(self._num_workers)]

    async def _worker_loop(self, instance):
        logging.logger.debug('Starting worker {}/{} on {} queue'.format(str(instance), str(self._num_workers), self._name))
        while True:
            new_msg = False
            try:
                item = await self._queue.get()
                new_msg = True
                if item.__class__ is ExitMessage:
                    logging.logger.debug('Stopping worker {}/{} on {} queue'.format(str(instance), str(self._num_workers),self._name))
                    break
                args, kwargs = item
                await asyncio.wait_for(self._on_msg(*args, **kwargs),self._timeout, loop=self._loop)
            except (KeyboardInterrupt, MemoryError, SystemExit):
                ex_info=traceback.format_exc().splitlines()
                for line in ex_info:
                    logging.logger.error(line)
                raise
            except BaseException:
                ex_info=traceback.format_exc().splitlines()
                for line in ex_info:
                    logging.logger.error(line)
            finally:
                if new_msg:
                    self._queue.task_done()

    async def join(self):
        if not self._workers:
            return
        for _ in range(self._num_workers):
            await self._queue.put(ExitMessage())
        try:
            await asyncio.gather(*self._workers, loop=self._loop)
            self._workers = None
        except:
            ex_info=traceback.format_exc().splitlines()
            for line in ex_info:
                logging.logger.error(line)
            raise
        finally:
            logging.logger.debug('Exiting AsyncQueue {}'.format(self._name))

    async def push(self, *args,**kwargs):
        await self._queue.put((args,kwargs))
        return None

